-module(rumBkdConsole).
%% Configuration is a proplist with the following keys:
%%`level'            - log level to use
%%`use_stderr'       - either `true' or `false', defaults to false. If set to true, use standard error to output console log messages
%%`formatter'        - the module to use when formatting log messages. Defaults to `rumFormatter'
%%`formatter_config' - the format configuration string. Defaults to `time [ severity ] message'

-behaviour(gen_emm).

-include("rumDef.hrl").

-define(TERSE_FORMAT, [time, " ", color, "[", severity, "] ", message]).
-define(RumDefConsoleFmtCfg, ?TERSE_FORMAT ++ [eol()]).
-define(RumDefConsoleOpts, [{use_stderr, false}, {group_leader, false}, {id, ?MODULE}, {formatter, rumFormatter}, {formatter_config, ?RumDefConsoleFmtCfg}]).

-export([
   init/1
   , handleCall/2
   , handleEvent/2
   , handleInfo/2
   , terminate/2
   , code_change/3
]).

-record(state, {
   id :: atom() | {atom(), any()}
   , level :: rumMaskLevel()
   , out = user :: user | standard_error | pid()
   , formatter :: atom()
   , formatCfg :: any()
   , colors = [] :: list()
}).

-spec init([rumConsoleOpt(), ...]) -> {ok, #state{}} | {error, atom()}.
init(Opts) ->
   case isNewStyleConsole() of
      false ->
         Msg = "Lager's console backend is incompatible with the 'old' shell, not enabling it",
         %% be as noisy as possible, log to every possible place
         try alarm_handler:set_alarm({?MODULE, "WARNING: " ++ Msg})
         catch
            _:_ -> error_logger:warning_msg(Msg ++ "~n")
         end,
         io:format("WARNING: " ++ Msg ++ "~n"),
         ?INT_LOG(warning, Msg, []),
         {error, {fatal, old_shell}};
      _ ->
         true = checkOpts(Opts),
         Colors = ?IIF(rumUtil:get_env(colored, false), rumUtil:get_env(colors, []), []),
         Level = rumUtil:get_opt(level, Opts, undefined),
         LevelMask = rumUtil:configToMask(Level),
         [UseErr, GroupLeader, Id, Formatter, Config] = [rumUtil:get_opt(Key, Opts, Def) || {Key, Def} <- ?RumDefConsoleOpts],
         Out = ?IIF(UseErr, standard_error, ?IIF(GroupLeader == false, user, begin erlang:monitor(process, GroupLeader), GroupLeader end)),
         {ok, #state{level = LevelMask, id = Id, out = Out, formatter = Formatter, formatCfg = Config, colors = Colors}}
   end.

checkOpts([]) -> true;
checkOpts([{id, {?MODULE, _}} | T]) ->
   checkOpts(T);
checkOpts([{level, Level} | T]) ->
   ?IIF(rumUtil:validateLogLevel(Level) /= false, checkOpts(T), {error, {bad_level, Level}});
checkOpts([{use_stderr, Flag} | T]) when is_boolean(Flag) ->
   checkOpts(T);
checkOpts([{formatter, M} | T]) when is_atom(M) ->
   checkOpts(T);
checkOpts([{formatter_config, C} | T]) when is_list(C) ->
   checkOpts(T);
checkOpts([{group_leader, L} | T]) when is_pid(L) ->
   checkOpts(T);
checkOpts([H | _]) ->
   {error, {invalid_opt, H}}.

handleCall(mGetLogLevel, State) ->
   {reply, State#state.level, State};
handleCall({mSetLogLevel, Level}, State) ->
   case rumUtil:validateLogLevel(Level) of
      false ->
         {reply, {error, bad_loglevel}, State};
      LevelMask ->
         {reply, ok, State#state{level = LevelMask}}
   end;
handleCall(_Msg, State) ->
   ?ERR("~p call receive unexpect msg ~p ~n ", [?MODULE, _Msg]),
   {reply, ok, State}.

handleEvent({mWriteLog, Message}, #state{level = Level, out = Out, formatter = Formatter, formatCfg = FormatConfig, colors = Colors, id = ID}) ->
   case rumUtil:isLoggAble(Message, Level, ID) of
      true ->
         io:put_chars(Out, Formatter:format(Message, FormatConfig, Colors)),
         kpS;
      _ ->
         kpS
   end;
handleEvent(_Msg, _State) ->
   ?ERR("~p event receive unexpect msg ~p ~n ", [?MODULE, _Msg]),
   kpS.

handleInfo({'DOWN', _, process, Out, _}, #state{out = Out}) ->
   removeEpm;
handleInfo(_Msg, _State) ->
   ?ERR("~p info receive unexpect msg ~p ~n", [?MODULE, _Msg]),
   kpS.

terminate(removeEpm, State) ->
   %% have to do this asynchronously because we're in the event handlr
   spawn(fun() -> eRum:clear_trace_by_destination(State#state.id) end),
   ok;
terminate(_Reason, _State) ->
   ok.

code_change(_OldVsn, State, _Extra) ->
   {ok, State}.

eol() ->
   ?IIF(rumUtil:get_env(colored, false), "\e[0m\r\n", "\r\n").

isNewStyleConsole() ->
   %% Criteria:
   %% 1. If the user has specified '-noshell' on the command line,
   %%    then we will pretend that the new-style console is available.
   %%    If there is no shell at all, then we don't have to worry
   %%    about log events being blocked by the old-style shell.
   %% 2. Windows doesn't support the new shell, so all windows users
   %%    have is the oldshell.
   %% 3. If the user_drv process is registered, all is OK.
   %%    'user_drv' is a registered proc name used by the "new"
   %%    console driver.
   init:get_argument(noshell) /= error orelse element(1, os:type()) /= win32 orelse is_pid(whereis(user_drv)).
